@@ -3,6 +3,8 @@ package ast
 import (
 	"strings"
 
+	"github.com/evanw/esbuild/internal/sourcemap"
+
 	"github.com/evanw/esbuild/internal/compat"
 )
 
@@ -248,6 +250,11 @@ type LocRef struct {
 	Ref Ref
 }
 
+type Span struct {
+	Text  string
+	Range Range
+}
+
 // This is used to represent both file system paths (IsAbsolute == true) and
 // abstract module paths (IsAbsolute == false). Abstract module paths represent
 // "virtual modules" when used for an input file and "package paths" when used
@@ -1273,6 +1280,9 @@ type AST struct {
 	NamedExports            map[string]Ref
 	TopLevelSymbolToParts   map[Ref][]uint32
 	ExportStarImportRecords []uint32
+
+	SourceMapComment Span
+	SourceMap        *sourcemap.SourceMap
 }
 
 func (ast *AST) HasCommonJSFeatures() bool {
