@@ -236,52 +236,118 @@ func parseFile(args parseArgs) {
 			fmt.Sprintf("File extension not supported: %s", args.prettyPath))
 	}
 
+	// Stop now if parsing failed
+	if !result.ok {
+		args.results <- result
+		return
+	}
+
 	// Run the resolver on the parse thread so it's not run on the main thread.
 	// That way the main thread isn't blocked if the resolver takes a while.
-	if result.ok && args.options.IsBundling {
+	if args.options.IsBundling {
 		result.resolveResults = make([]*resolver.ResolveResult, len(result.file.ast.ImportRecords))
 
-		// Resolve relative to the parent directory of the source file with the
-		// import path. Just use the current directory if the source file is virtual.
-		var sourceDir string
-		if source.KeyPath.IsAbsolute {
-			sourceDir = args.fs.Dir(source.KeyPath.Text)
-		} else if args.absResolveDir != "" {
-			sourceDir = args.absResolveDir
-		} else {
-			sourceDir = args.fs.Cwd()
-		}
+		if len(result.file.ast.ImportRecords) > 0 {
+			cache := make(map[string]*resolver.ResolveResult)
 
-		for _, part := range result.file.ast.Parts {
-			for _, importRecordIndex := range part.ImportRecordIndices {
-				// Don't try to resolve imports that are already resolved
-				record := &result.file.ast.ImportRecords[importRecordIndex]
-				if record.SourceIndex != nil {
-					continue
-				}
+			// Resolve relative to the parent directory of the source file with the
+			// import path. Just use the current directory if the source file is virtual.
+			var sourceDir string
+			if source.KeyPath.IsAbsolute {
+				sourceDir = args.fs.Dir(source.KeyPath.Text)
+			} else if args.absResolveDir != "" {
+				sourceDir = args.absResolveDir
+			} else {
+				sourceDir = args.fs.Cwd()
+			}
 
-				// Run the resolver and log an error if the path couldn't be resolved
-				resolveResult := args.res.Resolve(sourceDir, record.Path.Text)
-				if resolveResult == nil {
-					// Failed imports inside a try/catch are silently turned into
-					// external imports instead of causing errors. This matches a common
-					// code pattern for conditionally importing a module with a graceful
-					// fallback.
-					if !record.IsInsideTryBody {
-						r := source.RangeOfString(record.Loc)
-						args.log.AddRangeError(&source, r, fmt.Sprintf("Could not resolve %q", record.Path.Text))
+			for _, part := range result.file.ast.Parts {
+				for _, importRecordIndex := range part.ImportRecordIndices {
+					// Don't try to resolve imports that are already resolved
+					record := &result.file.ast.ImportRecords[importRecordIndex]
+					if record.SourceIndex != nil {
+						continue
 					}
-					continue
-				}
 
-				result.resolveResults[importRecordIndex] = resolveResult
+					// Cache the path in case it's imported multiple times in this file
+					if resolveResult, ok := cache[record.Path.Text]; ok {
+						result.resolveResults[importRecordIndex] = resolveResult
+						continue
+					}
+
+					// Run the resolver and log an error if the path couldn't be resolved
+					resolveResult := args.res.Resolve(sourceDir, record.Path.Text)
+					cache[record.Path.Text] = resolveResult
+
+					if resolveResult == nil {
+						// Failed imports inside a try/catch are silently turned into
+						// external imports instead of causing errors. This matches a common
+						// code pattern for conditionally importing a module with a graceful
+						// fallback.
+						if !record.IsInsideTryBody {
+							r := source.RangeOfString(record.Loc)
+							args.log.AddRangeError(&source, r, fmt.Sprintf("Could not resolve %q", record.Path.Text))
+						}
+						continue
+					}
+
+					result.resolveResults[importRecordIndex] = resolveResult
+				}
 			}
 		}
 	}
 
+	// Attempt to parse the source map if present
+	if args.options.SourceMap != config.SourceMapNone && result.file.ast.SourceMapComment.Text != "" {
+		if path, contents := extractSourceMapFromComment(args.log, args.fs, &source, result.file.ast.SourceMapComment); contents != nil {
+			prettyPath := path.Text
+			if path.IsAbsolute {
+				prettyPath = args.res.PrettyPath(prettyPath)
+			}
+			result.file.ast.SourceMap = parser.ParseSourceMap(args.log, logging.Source{
+				KeyPath:    path,
+				PrettyPath: prettyPath,
+				Contents:   *contents,
+			})
+		}
+	}
+
 	args.results <- result
 }
 
+func extractSourceMapFromComment(log logging.Log, fs fs.FS, source *logging.Source, comment ast.Span) (ast.Path, *string) {
+	// Data URL
+	if strings.HasPrefix(comment.Text, "data:") {
+		if strings.HasPrefix(comment.Text, "data:application/json;base64,") {
+			n := int32(len("data:application/json;base64,"))
+			encoded := comment.Text[n:]
+			decoded, err := base64.StdEncoding.DecodeString(encoded)
+			if err != nil {
+				r := ast.Range{Loc: ast.Loc{Start: comment.Range.Loc.Start + n}, Len: comment.Range.Len - n}
+				log.AddRangeWarning(source, r, "Invalid base64 data in source map")
+				return ast.Path{}, nil
+			}
+			contents := string(decoded)
+			return ast.Path{Text: "sourceMappingURL in " + source.PrettyPath}, &contents
+		}
+	}
+
+	// Absolute path
+	if source.KeyPath.IsAbsolute {
+		absPath := fs.Join(fs.Dir(source.KeyPath.Text), comment.Text)
+		contents, ok := fs.ReadFile(absPath)
+		if !ok {
+			log.AddRangeWarning(source, comment.Range, fmt.Sprintf("Could not find source map file: %s", absPath))
+			return ast.Path{}, nil
+		}
+		return ast.Path{IsAbsolute: true, Text: absPath}, &contents
+	}
+
+	// Anything else is unsupported
+	log.AddRangeWarning(source, comment.Range, "Unsupported source map comment")
+	return ast.Path{}, nil
+}
+
 func loaderFromFileExtension(extensionToLoader map[string]config.Loader, base string) config.Loader {
 	// Pick the loader with the longest matching extension. So if there's an
 	// extension for ".css" and for ".module.css", we want to match the one for
@@ -564,11 +630,6 @@ type compileResult struct {
 	// This is the line and column offset since the previous JavaScript string
 	// or the start of the file if this is the first JavaScript string.
 	generatedOffset lineColumnOffset
-
-	// The source map contains the original source code, which is quoted in
-	// parallel for speed. This is only filled in if the SourceMap option is
-	// enabled.
-	quotedSource string
 }
 
 func (b *Bundle) Compile(log logging.Log, options config.Options) []OutputFile {
@@ -582,14 +643,7 @@ func (b *Bundle) Compile(log logging.Log, options config.Options) []OutputFile {
 	}
 
 	// Determine the lowest common ancestor of all entry points
-	entryPointAbsPaths := make([]string, 0, len(b.entryPoints))
-	for _, entryPoint := range b.entryPoints {
-		keyPath := b.sources[entryPoint].KeyPath
-		if keyPath.IsAbsolute {
-			entryPointAbsPaths = append(entryPointAbsPaths, keyPath.Text)
-		}
-	}
-	lcaAbsPath := lowestCommonAncestorDirectory(b.fs, entryPointAbsPaths)
+	lcaAbsPath := b.lowestCommonAncestorDirectory(options.CodeSplitting)
 
 	type linkGroup struct {
 		outputFiles    []OutputFile
@@ -675,15 +729,43 @@ func (b *Bundle) Compile(log logging.Log, options config.Options) []OutputFile {
 	return outputFiles
 }
 
-func lowestCommonAncestorDirectory(fs fs.FS, absPaths []string) string {
+func (b *Bundle) lowestCommonAncestorDirectory(codeSplitting bool) string {
+	isEntryPoint := make(map[uint32]bool)
+	for _, entryPoint := range b.entryPoints {
+		isEntryPoint[entryPoint] = true
+	}
+
+	// If code splitting is enabled, also treat dynamic imports as entry points
+	if codeSplitting {
+		for _, sourceIndex := range findReachableFiles(b.sources, b.files, b.entryPoints) {
+			file := b.files[sourceIndex]
+			for _, part := range file.ast.Parts {
+				for _, importRecordIndex := range part.ImportRecordIndices {
+					if record := &file.ast.ImportRecords[importRecordIndex]; record.SourceIndex != nil && record.Kind == ast.ImportDynamic {
+						isEntryPoint[*record.SourceIndex] = true
+					}
+				}
+			}
+		}
+	}
+
+	// Ignore any paths for virtual modules (that don't exist on the file system)
+	absPaths := make([]string, 0, len(isEntryPoint))
+	for entryPoint := range isEntryPoint {
+		keyPath := b.sources[entryPoint].KeyPath
+		if keyPath.IsAbsolute {
+			absPaths = append(absPaths, keyPath.Text)
+		}
+	}
+
 	if len(absPaths) == 0 {
 		return ""
 	}
 
-	lowestAbsDir := fs.Dir(absPaths[0])
+	lowestAbsDir := b.fs.Dir(absPaths[0])
 
 	for _, absPath := range absPaths[1:] {
-		absDir := fs.Dir(absPath)
+		absDir := b.fs.Dir(absPath)
 		lastSlash := 0
 		a := 0
 		b := 0
