@@ -247,6 +247,9 @@ type Lexer struct {
 	codePoint                       rune
 	StringLiteral                   []uint16
 	Identifier                      string
+	JSXFactoryPragmaComment         ast.Span
+	JSXFragmentPragmaComment        ast.Span
+	SourceMappingURL                ast.Span
 	Number                          float64
 	rescanCloseBraceAsTemplateToken bool
 	json                            json
@@ -2218,6 +2221,74 @@ func (lexer *Lexer) addRangeError(r ast.Range, text string) {
 	}
 }
 
+func hasPrefixWithWordBoundary(text string, prefix string) bool {
+	t := len(text)
+	p := len(prefix)
+	if t >= p && text[0:p] == prefix {
+		if t == p {
+			return true
+		}
+		c, _ := utf8.DecodeRuneInString(text[p:])
+		if !IsIdentifierContinue(c) {
+			return true
+		}
+	}
+	return false
+}
+
+type pragmaArg uint8
+
+const (
+	pragmaNoSpaceFirst pragmaArg = iota
+	pragmaSkipSpaceFirst
+)
+
+func scanForPragmaArg(kind pragmaArg, start int, pragma string, text string) (ast.Span, bool) {
+	text = text[len(pragma):]
+	start += len(pragma)
+
+	if text == "" {
+		return ast.Span{}, false
+	}
+
+	// One or more whitespace characters
+	c, width := utf8.DecodeRuneInString(text)
+	if kind == pragmaSkipSpaceFirst {
+		if !IsWhitespace(c) {
+			return ast.Span{}, false
+		}
+		for IsWhitespace(c) {
+			text = text[width:]
+			start += width
+			if text == "" {
+				return ast.Span{}, false
+			}
+			c, width = utf8.DecodeRuneInString(text)
+		}
+	}
+
+	// One or more non-whitespace characters
+	i := 0
+	for !IsWhitespace(c) {
+		i += width
+		if i >= len(text) {
+			break
+		}
+		c, width = utf8.DecodeRuneInString(text[i:])
+		if IsWhitespace(c) {
+			break
+		}
+	}
+
+	return ast.Span{
+		Text: text[:i],
+		Range: ast.Range{
+			Loc: ast.Loc{Start: int32(start)},
+			Len: int32(i),
+		},
+	}, true
+}
+
 func (lexer *Lexer) scanCommentText() {
 	text := lexer.source.Contents[lexer.start:lexer.end]
 	hasPreserveAnnotation := len(text) > 2 && text[2] == '!'
@@ -2226,16 +2297,32 @@ func (lexer *Lexer) scanCommentText() {
 		switch text[i] {
 		case '#':
 			rest := text[i+1:]
-			if strings.HasPrefix(rest, "__PURE__") {
+			if hasPrefixWithWordBoundary(rest, "__PURE__") {
 				lexer.HasPureCommentBefore = true
+			} else if strings.HasPrefix(rest, " sourceMappingURL=") {
+				if arg, ok := scanForPragmaArg(pragmaNoSpaceFirst, lexer.start+i+1, " sourceMappingURL=", rest); ok {
+					lexer.SourceMappingURL = arg
+				}
 			}
 
 		case '@':
 			rest := text[i+1:]
-			if strings.HasPrefix(rest, "__PURE__") {
+			if hasPrefixWithWordBoundary(rest, "__PURE__") {
 				lexer.HasPureCommentBefore = true
-			} else if strings.HasPrefix(rest, "preserve") || strings.HasPrefix(rest, "license") {
+			} else if hasPrefixWithWordBoundary(rest, "preserve") || hasPrefixWithWordBoundary(rest, "license") {
 				hasPreserveAnnotation = true
+			} else if hasPrefixWithWordBoundary(rest, "jsx") {
+				if arg, ok := scanForPragmaArg(pragmaSkipSpaceFirst, lexer.start+i+1, "jsx", rest); ok {
+					lexer.JSXFactoryPragmaComment = arg
+				}
+			} else if hasPrefixWithWordBoundary(rest, "jsxFrag") {
+				if arg, ok := scanForPragmaArg(pragmaSkipSpaceFirst, lexer.start+i+1, "jsxFrag", rest); ok {
+					lexer.JSXFragmentPragmaComment = arg
+				}
+			} else if strings.HasPrefix(rest, " sourceMappingURL=") {
+				if arg, ok := scanForPragmaArg(pragmaNoSpaceFirst, lexer.start+i+1, " sourceMappingURL=", rest); ok {
+					lexer.SourceMappingURL = arg
+				}
 			}
 		}
 	}
